# pie_tree.py
# by Nick Gerend

import pandas as pd
import numpy as np
from math import pi, cos, sin, sqrt
import random
import matplotlib.pyplot as plt
import pkg_resources

def load_aircraft_df():
    # data sourced from https://www.faa.gov/licenses_certificates/aircraft_certification/aircraft_registry/releasable_aircraft_download/
    stream = pkg_resources.resource_stream(__name__, 'data/aircraft_2020.csv')
    return pd.read_csv(stream, dtype=pd.StringDtype())

class point:
    def __init__(self, index, level, group, x, y, path, count): 
        self.index = index
        self.level = level
        self.group = group
        self.x = x
        self.y = y
        self.path = path
        self.count = count
    def to_dict(self):
        return {
            'index' : self.index,
            'level' : self.level,
            'group' : self.group,
            'x' : self.x,
            'y' : self.y,
            'path' : self.path,
            'count' : self.count }

def dict_start_end(group, start, count):
    df_dict = group.size()
    df_dict = df_dict.to_dict()
    #start = 0
    end = 0
    count_ = 0
    for key in df_dict:
        count_ = df_dict[key]
        end = df_dict[key] + start
        df_dict[key] = []
        df_dict[key].append(end/count - start/count)
        df_dict[key].append(0.) # r1
        df_dict[key].append(0.) # r2
        df_dict[key].append(0.) # a1
        df_dict[key].append(0.) # a2
        df_dict[key].append(0.) # jump
        df_dict[key].append(count_) # count
        start = end
    return df_dict

def inner_rad(area, r1, f):
    num = area/f+pi*r1**2
    return sqrt(num/pi)

def inner_angle(area, r1, r2, fill_angle):
    num = area/(pi*r2**2-pi*r1**2)
    return num*fill_angle

def location(r1, r2, a1, a2, area, orientation='v', fill_angle=360.):
    area_o = pi*r2**2-pi*r1**2
    out = 0.
    if orientation == 'h':
        f = (a2-a1)/fill_angle
        r3 = sqrt(((area/(f))+pi*r1**2)/pi)
        out = r1, r3, a1, a2 
    else:
        a3 = (area*fill_angle)/(pi*r2**2-pi*r1**2)
        out = r1, r2, a1, a3
    return out

def pie_tree_calc(df, groupers, r1, r2, start_angle, end_angle, points, default_sort = False, default_sort_override = True, default_sort_override_reversed = False, all_vertical = False):

    #region algorithm

    #region initialize
    list_group = []
    for i in range(len(groupers)):
        group_i = groupers[0:i+1]
        df_group = df.groupby(group_i, sort=default_sort)
        df_dict = dict_start_end(df_group, 0., len(df))
        if default_sort_override:
            df_dict = dict(sorted(df_dict.items(), key=lambda item: item[1][0], reverse=default_sort_override_reversed))
        list_group.append(df_dict)
    last = 0.
    for i in list_group[0]:
        list_group[0][i][1] = r1
        list_group[0][i][2] = r2
        list_group[0][i][3] = last * (end_angle-start_angle) + start_angle
        list_group[0][i][4] = (last + list_group[0][i][0]) * (end_angle-start_angle) + start_angle
        last += list_group[0][i][0]
    #endregion

    #region point grid
    area = (pi*r2**2-pi*r1**2)*((end_angle-start_angle)/360.)
    for i in range(1, len(groupers), 1):
        ri = 0
        ai = 0
        rc1, rc2, ac1, ac2 = 0., 0., 0., 0.
        for j in list_group[i]:
            # parent box
            key = j[0:i]
            if i == 1:
                key = key[0]
            # key = ','.join(key)
            rp1 = list_group[i-1][key][1]
            rp2 = list_group[i-1][key][2]
            ap1 = list_group[i-1][key][3]
            ap2 = list_group[i-1][key][4]
            jump = list_group[i-1][key][5]
            
            # child box
            areac = list_group[i][j][0]*area
            if i % 2 == 0 or all_vertical:
                rc1, rc2, ac1, ac2 = location(rp1, rp2, ap1, ap2, areac, 'v')
                if key == ('Individual', 'Fixed wing single engine'):
                    stop = 1
                list_group[i-1][key][5] += ac2
                ac1 += jump
                ac2 += ac1
            else:
                if jump == 0:
                    jump = rp1
                rc1, rc2, ac1, ac2 = location(jump, rp2, ap1, ap2, areac, 'h')
                if key == 'Non Citizen Co-Owned':
                    stop = 1
                list_group[i-1][key][5] = rc2
            list_group[i][j][1] = rc1
            list_group[i][j][2] = rc2
            list_group[i][j][3] = ac1
            list_group[i][j][4] = ac2
    #endregion

    #region draw
    ix = 0
    level = 1
    list_xy = []
    for i in range(len(groupers)):
        for j in list_group[i]:
            ad1 = list_group[i][j][3]
            ad2 = list_group[i][j][4]
            a1 = (ad1-90.)*pi/180.
            a2 = (ad2-90.)*pi/180.
            r_1 = list_group[i][j][1]
            r_2 = list_group[i][j][2]
            x1 = r_1*cos(a1)
            y1 = r_1*sin(a1)

            point_frac = (ad2-ad1)/365.
            points1 = int(point_frac*points)+5
            points2 = int(point_frac*points*r_1/r_2)+5
            angles1 = np.linspace(a1, a2, num=points1)
            angles2 = np.linspace(a2, a1, num=points2)

            count = list_group[i][j][6]

            list_xy.append(point(ix, level, j, x1, -y1, 0, count))
            for k in range(len(angles1)):
                list_xy.append(point(ix, level, j, r_2*cos(angles1[k]), -r_2*sin(angles1[k]), k+1, count))
            for k in range(len(angles2)):
                list_xy.append(point(ix, level, j, r_1*cos(angles2[k]), -r_1*sin(angles2[k]), len(angles1)+k+1, count))
            
        level += 1
    #endregion

    #endregion

    df_out = pd.DataFrame.from_records([s.to_dict() for s in list_xy])
    return df_out

def pie_tree_plot(pie_tree_df, level = 1, transparency = 0.5, line_level = 0):
    fig, axs = plt.subplots()
    axs.set_aspect('equal', adjustable='box')
    df_lvl = pie_tree_df.loc[pie_tree_df['level'] == level]
    df_lvl_group = df_lvl.groupby(['group'])
    for group, rows in df_lvl_group:
        x = rows['x'].values
        y = rows['y'].values
        r = random.random()
        b = random.random()
        g = random.random()
        color = (r, g, b)
        axs.fill(x, y, alpha=transparency, fc=color)
        plt.plot(x, y, 'k-', linewidth=0.5)
    if line_level > 0:
        df_lvl = pie_tree_df.loc[pie_tree_df['level'] == line_level]
        df_lvl_group = df_lvl.groupby(['group'])
        for group, rows in df_lvl_group:
            x = rows['x'].values
            y = rows['y'].values
            plt.plot(x, y, 'k-', linewidth=3)
    plt.show()


# multi_chord.py
# by Nick Gerend

import pandas as pd
import numpy as np
from math import sqrt, sin, cos, atan2, pi, isnan
import matplotlib.pyplot as plt
import random
import pkg_resources

def load_olympics_df():
    # data sourced from https://www.kaggle.com/heesoo37/olympic-history-data-a-thorough-analysis/data
    stream = pkg_resources.resource_stream(__name__, 'data/athlete_events.csv')
    return pd.read_csv(stream)

class group:
    def __init__(self, group, element, value, position, offset, x1, x2): 
        self.group = group
        self.element = element
        self.value = value
        self.position = position
        self.offset = offset
        self.x1 = x1
        self.x2 = x2
    def to_dict(self):
        return {
            'group' : self.group,
            'element' : self.element,
            'value' : self.value,
            'position' : self.position,
            'offset' : self.offset,
            'x1' : self.x1, 
            'x2' : self.x2 }

class point:
    def __init__(self, group, value, count, x, y, path, type='chord'): 
        self.group = group
        self.value = value
        self.count = count
        self.x = x
        self.y = y
        self.path = path
        self.type = type
    def to_dict(self):
        return {
            'group' : self.group,
            'value' : self.value,
            'count' : self.count,
            'x' : self.x,
            'y' : self.y,
            'path' : self.path,
            'type' : self.type }

def polarize(x, max_x, y_offset = 0.):
    y = max_x/4.+y_offset
    angle = (2.*pi)*(((x)%(max_x))/(max_x))
    angle_deg = angle * 180./pi
    angle_rotated = (abs(angle_deg-360.)+90.) % 360. 
    angle_new = angle_rotated * pi/180.
    x_out = (y)*cos(angle_new)
    y_out = (y)*sin(angle_new)

    #x_out, y_out = rink_scale(x_out, y_out, 1., 1.) #2.19608

    return x_out, y_out

def LnToPntDst(x0, y0, x1, y1, x2, y2):
    n = abs((y1-y2)*x0+(x2-x1)*y0+x1*y2-x2*y1)
    d = sqrt((x2-x1)**2+(y2-y1)**2)
    return n/d

def DistBtwTwoPnts(x1, y1, x2, y2):
    return sqrt((x2-x1)**2+(y2-y1)**2)

def Rotate(x, y, angledeg, x_offset, y_offset):
    xa = x*cos(angledeg*pi/180) + y*sin(angledeg*pi/180)
    ya = -x*sin(angledeg*pi/180) + y*cos(angledeg*pi/180)
    xa -= x_offset
    ya -= y_offset
    return xa, ya

def AngleByTwoPnts(x1, y1, x2, y2):
    return atan2(x2-x1, y2-y1)*180/pi - 90

def chord(x0, y0, x1, y1, x2, y2, points, h_override=0.):
    h = LnToPntDst(x0, y0, x1, y1, x2, y2)
    w = DistBtwTwoPnts(x1, y1, x2, y2)

    if h_override == 0.:
        new_h = (1.-(h/w)/10.)*h
        if new_h < h*0.01:
            h = h*0.01
        else:
            h = new_h
    else:
        h = h*h_override

    a = AngleByTwoPnts(x1, y1, x2, y2)
    xr = []
    yr = []
    for i in range(points+1):
        arc_percent = i/(points/2.)
        if i > points/2.:
            arc_percent = (points-i)/(points/2.)
        if i == 0 or i == points:
            arc = 0.
        else:
            arc = sqrt((h/2.)**2-((h/2.)-(h/2.)/((points)/2.)*i)**2.)
        percent = arc/(h/2.)

        y_1 = -percent*arc+(1-percent)*arc_percent
        y_2 = percent*arc+(1-percent)*arc_percent
        xr_1, yr_1 = Rotate(i/points*w, y_1, a, -x1, -y1)
        xr_2, yr_2 = Rotate(i/points*w, y_2, a, -x1, -y1)

        d1 =  DistBtwTwoPnts(x0, y0, xr_1, yr_1)
        d2 =  DistBtwTwoPnts(x0, y0, xr_2, yr_2)

        if d1 < d2:
            xr.append(xr_1)
            yr.append(yr_1)
        else:
            xr.append(xr_2)
            yr.append(yr_2)
    return list(zip(xr, yr))
    
def rescale(x, xmin, xmax, newmin, newmax):
    rescaled = (newmax-newmin)*((x-xmin)/(xmax-xmin))+newmin
    return rescaled

def multi_chord_agg(df, order, percent = 100., buffer = 1., elements_offset = 0.04, elements_height = 0.04, group_offset = 0., group_height = 0.04):

    #region initialize element list
    element_list = []
    for i, row in df.iterrows():
        group_i = row['group']
        value_i = row['value']
        elements = group_i.split(',')
        for j in range(len(elements)):
            element_list.append(group(group_i, elements[j], value_i, 0, 0., 0., 0.))
    df_elements = pd.DataFrame.from_records([s.to_dict() for s in element_list])
    #endregion

    #region set up group propperties
    df_group = df_elements.groupby(['element'])['value'].sum().reset_index()
    buffer = buffer/100.*df_group['value'].sum()
    df_group['offset'] = 0.

    order_list = order.split(',')
    df_group['order'] = [order_list.index(x)+1 for x in df_group['element']]
    
    offset = 0.
    for i, row in df_group.sort_values(by=['order'], ascending=True).iterrows():
        df_group.at[i,'offset'] = offset
        offset += row['value']
    dict_group = df_group.set_index('element').T.to_dict('list')
    #endregion

    #region endpoints
    value_offset = 0.
    last_group = order_list[0]
    df_elements['order'] = [order_list.index(x)+1 for x in df_elements['element']]
    for i, row in df_elements.sort_values(by=['order', 'value'], ascending=True).iterrows():
        if row['element'] != last_group:
            value_offset = 0.
            last_group = row['element']
        
        order_i = row['order']
        offset = dict_group[row['element']][1]

        df_elements.at[i,'offset'] = value_offset
        df_elements.at[i,'x1'] = offset + value_offset + buffer*(order_i-1)
        df_elements.at[i,'x2'] = offset + value_offset + row['value'] + buffer*(order_i-1)

        value_offset += row['value']
    #endregion

    #region setup virtual points
    span = df_elements['x2'].max()+buffer

    if percent < 100.:
        span = df_elements['x2'].max()*(1./(percent/100.))

    df_elements_prior = df_elements.copy()
    df_elements_next = df_elements.copy()
    df_elements_prior['x1'] = df_elements_prior['x1'] - span
    df_elements_prior['x2'] = df_elements_prior['x2'] - span
    df_elements_prior['position'] = -1
    df_elements_next['x1'] = df_elements_next['x1'] + span
    df_elements_next['x2'] = df_elements_next['x2'] + span
    df_elements_next['position'] = 1
    dfs = [df_elements_prior, df_elements, df_elements_next]
    df_combined = pd.concat(dfs, axis=0)
    #endregion

    #region point dictionary
    chord_dict = {}
    for name, group_i in df_combined.groupby(['group', 'position']):
        x_list = []
        key = name
        for i, row in group_i.sort_values(by=['order'], ascending=True).iterrows():
            e = row['element']
            x1 = row['x1']
            x2 = row['x2']
            x_list.append([e,x1,x2])
        chord_dict[key] = x_list
    #endregion

    #region select point path
    chord_arc_dict = {}
    for i, row in df.iterrows():
        list_i = chord_dict[row['group'],0]
        list_neg = chord_dict[row['group'],-1]
        list_pos = chord_dict[row['group'],1]

        if len(list_i) > 1:
            list_new = []
            p1 = list_i[0][2]
            p2 = 0.
            for j in range(len(list_i)):

                p1_0 = p1
                if j == len(list_i)-1:
                    j = -1

                p2_i = list_i[j+1][1]
                p2_neg = list_neg[j+1][1]
                p2_pos = list_pos[j+1][1]

                i_delta = abs(p1-p2_i)
                neg_delta = abs(p1-p2_neg)
                pos_delta = abs(p1-p2_pos)

                if i_delta < neg_delta:
                    if i_delta < pos_delta:
                        p2 = p2_i
                        p1 = list_i[j+1][2]
                    else:
                        p2 = p2_pos
                        p1 = list_pos[j+1][2]
                elif pos_delta < neg_delta:
                    p2 = p2_pos
                    p1 = list_pos[j+1][2]
                else:
                    p2 = p2_neg
                    p1 = list_neg[j+1][2]

                list_new.append([row['group'],p1_0, p2])
            chord_arc_dict[row['group']] = list_new
        else:
            chord_arc_dict[row['group']] = list_i
    #endregion

    #region draw point path
    list_xy = []
    for i, row in df.iterrows():
        group_i = row['group']
        value = row['value']
        count = len(group_i.split(','))
        list_path = chord_arc_dict[row['group']]

        if count == 1:
            #region 1 element
            p1 = list_path[0][1]
            p2 = list_path[0][2]
            points = int(abs(p2-p1)/(span)/2.*720)+10
            s1 = np.linspace(p2, p1, num=points)
            path = 1
            for k in range(len(s1)):
                x, y = polarize(s1[k], span)
                list_xy.append(point(group_i, value, count, x, y, path))
                path += 1
            x1, y1 = polarize(p1, span)
            x2, y2 = polarize(p2, span)
            chord_list = chord(0., 0., x1, y1, x2, y2, 20, 0.25)
            for k in range(len(chord_list)):
                list_xy.append(point(group_i, value, count, chord_list[k][0], chord_list[k][1], path))
                path += 1
            #endregion
        
        else:
            #region multiple elements

            #region first straight segment
            path = 1
            p1 = chord_dict[row['group'],0][0][1]
            p2 = list_path[0][1]
            points = int(abs(p2-p1)/(span)/2.*720)
            #straight between p1 to p2
            s1 = np.linspace(p1, p2, num=points)
            for k in range(len(s1)):
                    x, y = polarize(s1[k], span)
                    list_xy.append(point(group_i, value, count, x, y, path))
                    path += 1        
            #endregion
            
            for j in range(len(list_path)):

                #region curves
                p2 = list_path[j][1]
                x1, y1 = polarize(p2, span)
                p3 = list_path[j][2]
                x2, y2 = polarize(p3, span)
                chord_list = chord(0., 0., x1, y1, x2, y2, 50)
                for k in range(len(chord_list)):
                        list_xy.append(point(group_i, value, count, chord_list[k][0], chord_list[k][1], path))
                        path += 1
                #endregion

                #region lagging straight segments
                if j < len(list_path)-1:
                    p4 = list_path[j+1][1]
                    #straight between p3 to p4
                    points = int(abs(p4-p3)/(span)/2.*720)
                    s2 = np.linspace(p3, p4, num=points)
                    for k in range(len(s2)):
                        x, y = polarize(s2[k], span)
                        list_xy.append(point(group_i, value, count, x, y, path))
                        path += 1
                #endregion

            #endregion
    #endregion

    #region rescale to unit radius
    for i in range(len(list_xy)):
        list_xy[i].x = rescale(list_xy[i].x, -span/4., span/4., -1., 1.)
        list_xy[i].y = rescale(list_xy[i].y, -span/4., span/4., -1., 1.)
    #endregion

    #region add legends
    group_offset += elements_offset + elements_height
    for i, row in df_elements.iterrows():
        group_i =  row.group
        value = row.value
        count = len(group_i.split(','))
        element = row.element

        p1 = row.x1
        p2 = row.x2
        points = int(abs(p2-p1)/(span)/2.*720)+3
        
        path = 1
        top = np.linspace(p1, p2, num=points)
        x1 = 0
        y1 = 0
        initial = True
        for k in range(len(top)):
            top_k = rescale(top[k], -span/4., span/4., -1., 1.)
            x, y = polarize(top_k, 4., elements_offset)
            if initial:
                x1 = x
                y1 = y
                initial = False
            list_xy.append(point(group_i, value, count, x, y, path, 'group_'+element))
            path += 1
        bottom = np.linspace(p2, p1, num=points)
        for k in range(len(bottom)):
            bottom_k = rescale(bottom[k], -span/4., span/4., -1., 1.)
            x, y = polarize(bottom_k, 4., elements_offset + elements_height)
            list_xy.append(point(group_i, value, count, x, y, path, 'group_'+element))
            path += 1
        list_xy.append(point(group_i, value, count, x1, y1, path, 'group_'+element))

    for i, row in df_group.iterrows():
        group_i =  row.element
        value = row.value
        count = None

        p1 = row.offset + (row.order-1)*buffer
        p2 = row.offset + row.value + (row.order-1)*buffer
        points = int(abs(p2-p1)/(span)/2.*720)+3
        
        path = 1
        top = np.linspace(p1, p2, num=points)
        x1 = 0
        y1 = 0
        initial = True
        for k in range(len(top)):
            top_k = rescale(top[k], -span/4., span/4., -1., 1.)
            x, y = polarize(top_k, 4., group_offset)
            if initial:
                x1 = x
                y1 = y
                initial = False
            list_xy.append(point(group_i, value, count, x, y, path, 'element'))
            path += 1
        bottom = np.linspace(p2, p1, num=points)
        for k in range(len(bottom)):
            bottom_k = rescale(bottom[k], -span/4., span/4., -1., 1.)
            x, y = polarize(bottom_k, 4., group_offset + group_height)
            list_xy.append(point(group_i, value, count, x, y, path, 'element'))
            path += 1
        list_xy.append(point(group_i, value, count, x1, y1, path, 'element'))
    #endregion

    return list_xy

def multi_chord_groups(df, outer, inner, order=None):
    
    #region outer lookup
    df = df[[outer, inner]]
    group_count = df.groupby(outer)[inner].nunique().reset_index()
    outer_lookup = group_count.sort_values(by=[inner], ascending=False)
    outer_lookup['outer_ID']  = range(1, len(outer_lookup) + 1)
    outer_lookup['outer_ID'] = outer_lookup['outer_ID'].astype(str)
    outer_lookup = outer_lookup.rename({inner: 'e2_count'}, axis=1)
    #endregion

    #region check order:
    if order == None:
        order = ','.join(outer_lookup['outer_ID'].to_list())
    #endregion

    #region merge outer_ID to data
    df = pd.merge(df, outer_lookup, how='left', on=[outer])
    #endregion

    #region create chords
    e2_group = df.groupby(inner)
    chord_dict = {}
    for name, elements in e2_group:
        #collect ordered set of outer by e2
        key = ''
        for i, row in elements.sort_values(by=['outer_ID'], ascending=True).iterrows():
            outer = row['outer_ID']
            key += str(outer) + ','
        key = key[:-1]
        if key in chord_dict:
            chord_dict[key] += 1
        else:
            chord_dict[key] = 1

    df_chord = pd.Series(chord_dict, name='value')
    df_chord.index.name = 'group'
    df_chord = df_chord.reset_index()
    #endregion

    return df_chord, order

def multi_chord_alias(df, outer, inner, percent=100., order=None, buffer = 1., elements_offset = 0.04, elements_height = 0.04, group_offset = 0., group_height = 0.04):
    
    df_chord, order = multi_chord_groups(df, outer, inner, order)
    df_chord['value'] *= 100

    list_xy = multi_chord_agg(df_chord, order, percent, buffer, elements_offset, elements_height, group_offset, group_height)

    df_out = pd.DataFrame.from_records([s.to_dict() for s in list_xy])
    df_out['value'] /= 100

    return df_out

def multi_chord_on_groups_alias(df_chord, percent=100., order=None, buffer = 1., elements_offset = 0.04, elements_height = 0.04, group_offset = 0., group_height = 0.04):
    
    if order == None:
        order_dict = {}
        for i, row in df_chord.iterrows():
            items = row['group'].split(',')
            value = row['value']
            for j in range(len(items)):
                if items[j] in order_dict:
                    order_dict[items[j]] += value
                else:
                    order_dict[items[j]] = value
        df_order = pd.DataFrame(order_dict.items(), columns=['group', 'value'])
        order = df_order.sort_values(by=['value'], ascending=False)['group'].values.tolist()
        order = ",".join(order)
    
    df_chord['value'] *= 100

    list_xy = multi_chord_agg(df_chord, order, percent, buffer, elements_offset, elements_height, group_offset, group_height)

    df_out = pd.DataFrame.from_records([s.to_dict() for s in list_xy])
    df_out['value'] /= 100

    return df_out

def multi_chord_venn(df_mc):
    df_mc_cc = pd.concat([df_mc, df_mc['group'].str.get_dummies(sep=',')], axis = 1)
    df_mc_venn = pd.melt(df_mc_cc, id_vars=['group', 'value'], var_name='group2', value_name='value2')
    return df_mc_venn

def multi_chord_plot(df_mc, level = None, transparency = 0.5):
    max_count = int(max(df_mc['count'].unique()))
    colors = []
    for i in range(max_count):
        r = random.random()
        b = random.random()
        g = random.random()
        color = (r, g, b)
        colors.append(color)
    df_lvl_group = df_mc.groupby(['type','group'])
    fig, axs = plt.subplots()
    axs.set_aspect('equal', adjustable='box')
    for group, rows in df_lvl_group:
        x = rows['x'].values
        y = rows['y'].values
        count_check = rows['count'].values[0]
        color = ()
        set_linewidth=0.5
        if isinstance(count_check, (int, float)) and not isnan(count_check):
            color = colors[int(count_check)-1]
            if int(count_check) == level:
                set_linewidth=3
        else:
            r = random.random()
            b = random.random()
            g = random.random()
            color = (r, g, b)
        axs.fill(x, y, alpha=transparency, fc=color)
        plt.plot(x, y, 'k-', linewidth=set_linewidth)
    plt.show(block=True)

def multi_chord_get_alias(df, outer, inner):
    
    #region outer lookup
    df = df[[outer, inner]]
    group_count = df.groupby(outer)[inner].nunique().reset_index()
    outer_lookup = group_count.sort_values(by=[inner], ascending=False)
    outer_lookup['outer_ID']  = range(1, len(outer_lookup) + 1)
    outer_lookup['outer_ID'] = outer_lookup['outer_ID'].astype(str)
    outer_lookup = outer_lookup.rename({inner: 'e2_count'}, axis=1)
    #endregion

    return outer_lookup


import pathlib
import setuptools as st
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
setup(
    name="dataoutsider",
    version="1.1.1",
    description="Data visualization toolbox.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://pypi.org/project/dataoutsider/",
    author="Nick Gerend",
    author_email="nickgerend@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=st.find_packages(),
    install_requires=["pandas", "numpy", "matplotlib"],
    include_package_data=True,
    package_data={'': ['data/*.csv']},
)


# bubble_series.py
# by Nick Gerend

from math import sin, pi, cos, sqrt
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import random

class point:
    def __init__(self, item, x, y, A, group, path=0, xo=0, yo=0): 
        self.item = item
        self.x = x
        self.y = y
        self.A = A
        self.group = group
        self.path = path
        self.xo = xo
        self.yo = yo
    def to_dict(self):
        return {
            'item' : self.item,
            'x' : self.x,
            'y' : self.y,
            'A' : self.A,
            'group' : self.group,
            'path' : self.path,
            'xo' : self.xo,
            'yo' : self.yo }

def circle(id, xo, yo, r=1., points=50):
    if points % 2 != 0:
        points += 1
    list_ixyp = []
    angle = -90.
    path_i = 1
    for i in range(points+1):
        # id, x, y, path
        list_ixyp.append((id, xo+r*sin(angle*pi/180.), yo+r*cos(angle*pi/180.), path_i))
        angle += 1./points*360.
        path_i += 1
    list_ixyp.append((id, xo+r*sin(-90.*pi/180.), yo+r*cos(-90.*pi/180.), path_i))
    return list_ixyp

def circle_collide(x1, y1, x2, r1, r2, place='top'):
    y2 = None
    if (x1-r1 < x2-r2 and x1+r1 > x2-r2) or (x1-r1 < x2+r2 and x1+r1 > x2+r2):
        if place == 'bottom':
            y2 = y1 - sqrt(-x2**2 + 2*x2*x1 + r2**2 + 2*r2*r1 + r1**2 - x1**2)
        if place == 'top':
            y2 = sqrt(-x2**2 + 2*x2*x1 + r2**2 + 2*r2*r1 + r1**2 - x1**2) + y1
    return y2

def circle_collided(x1, y1, x2, y2, r1, r2):
    A = (x2-x1)**2 + (y2-y1)**2
    B = (r1+r2)**2
    result = False
    tol = 10
    if round(A, tol) < round(B, tol):
        result = True
    return result
        
def delete_multiple_element(list_object, indices):
    indices = sorted(indices, reverse=True)
    for idx in indices:
        if idx < len(list_object):
            list_object.pop(idx)

def bubble_group(df):
    dfi = df[['item', 'Min', 'Max']]
    records = dfi.to_records(index=False)
    list_c = list(records)
    group = 0
    groups = {}
    groups[group] = [list_c[0]]
    pop_list = []
    minx = list_c[0][1]
    maxx = list_c[0][2]
    list_c.pop(0)
    while len(list_c) > 0:
        for i in range(len(list_c)):
            minx2 = list_c[i][1]
            maxx2 = list_c[i][2]
            c1 = minx < minx2 and maxx > minx2
            c2 = minx < maxx2 and maxx > maxx2
            c3 = minx2 < minx and maxx2 > minx
            c4 = minx2 < maxx and maxx2 > maxx
            if c1 or c2 or c3 or c4:
                if minx2 < minx:
                    minx = minx2
                if maxx2 > maxx:
                    maxx = maxx2
                groups[group].append(list_c[i])
                pop_list.append(i)
        if len(pop_list) == 0:
            group += 1
            minx = list_c[0][1]
            maxx = list_c[0][2]
            groups[group] = [list_c[0]]
            list_c.pop(0)
        else:
            delete_multiple_element(list_c, pop_list)
            pop_list = []

    rows = []
    for key in groups:
        for i in range(len(groups[key])):
            rows.append((key,groups[key][i][0]))

    df_join = pd.DataFrame(rows, columns=['group', 'item'])
    df_out = df.merge(df_join, on=['item'], how='left')
    return df_out

def bubble_series_calc(df, item_col_name, position_col_name, size_col_name, r_buffer=0.0):
    df['item'] = df[item_col_name]
    df['x'] = df[position_col_name]
    df['A'] = df[size_col_name]
    df['Ao'] = df['A']
    df['r'] = np.sqrt(df['A']/pi) + r_buffer
    df['A'] = pi*df['r']**2
    df['Min'] = -df['r'] + df['x']
    df['Max'] = df['r'] + df['x']
    df = bubble_group(df)
    df_grouped = df.groupby(['group'])
    list_xy = []
    dict_groups = {}
    group_i = 0
    for group, rows in df_grouped:
        if group == 9:
            test = 'debug'
        rows_sort = rows.sort_values(by=['A'], ascending=False)
        for i, row in rows_sort.iterrows():
            item = row['item']
            x = row['x']
            r = row['r']
            A = row['A']
            y = 0.
            if group == 4 and group_i == 4:
                test = 'debug'
            if group_i == 0:
                dict_groups[group] = [(x,0.,r)]
                list_xy.append(point(item, x, 0., A, group))
            else:
                maxy = 0.
                miny = 0.
                yi = 0.
                for j in range(len(dict_groups[group])):
                    #try top and bottom of every prior-placed item
                    #place if no collision and minimum y-extent
                    xo = dict_groups[group][j][0]
                    yo = dict_groups[group][j][1]
                    ro = dict_groups[group][j][2]
                    yt = circle_collide(xo, yo, x, ro, r, place='top')
                    if yt == None:
                        if j == 0:
                            ylast = 0.
                        continue
                    yb = circle_collide(xo, yo, x, ro, r, place='bottom')
                    # check collion against every other placed item
                    ytest = [yt,yb]
                    collided = False
                    for k in range(len(dict_groups[group])):
                        xk = dict_groups[group][k][0]
                        yk = dict_groups[group][k][1]
                        rk = dict_groups[group][k][2]
                        if circle_collided(xk, yk, x, yt, rk, r):
                            ytest[0] = None
                        if circle_collided(xk, yk, x, yb, rk, r):
                            ytest[1] = None
                        if ytest[0] == None and ytest[1] == None:
                            collided = True
                            break
                    if collided:
                        continue
                    if ytest[0] == None:
                        yi = yb
                    elif ytest[1] == None:
                        yi = yt
                    else:
                        if abs(yt) > abs(yb):
                            yi = yb
                        else:
                            yi = yt
                    if j == 0:
                        ylast = yi
                        y = yi
                    elif abs(yi) < abs(ylast) or ylast == 0.:
                        y = yi
                    ylast = yi
                dict_groups[group].append((x,y,r))
                list_xy.append(point(item, x, y, A, group))
            ylast = 0.
            group_i += 1
        group_i = 0

    circles = []
    for i in range(len(list_xy)):
        item = list_xy[i].item
        x = list_xy[i].x
        y = list_xy[i].y
        A = list_xy[i].A
        group = list_xy[i].group
        r = sqrt(A/pi)
        r -= r_buffer
        A = pi*r**2
        circle_i = circle((item,group,A,x,y), x, y, r=r)
        circles.extend(circle_i)

    list_xyc = []
    for i in range(len(circles)):
        item = circles[i][0][0]
        x = circles[i][1]
        y = circles[i][2]
        A = circles[i][0][2]
        path = circles[i][3]
        group = circles[i][0][1]
        xo = circles[i][0][3]
        yo = circles[i][0][4]
        list_xyc.append(point(item, x, y, A, group, path, xo, yo))

    df_out = pd.DataFrame.from_records([s.to_dict() for s in list_xyc])
    return df_out

def bubble_series_plot(bubble_series_calc_df, transparency = 0.5):
    fig, axs = plt.subplots()
    axs.set_aspect('equal', adjustable='box')
    df_group = bubble_series_calc_df.groupby(['item'])
    for group, rows in df_group:
        x = rows['x'].values
        y = rows['y'].values
        r = random.random()
        b = random.random()
        g = random.random()
        color = (r, g, b)
        axs.fill(x, y, alpha=transparency, fc=color)
        plt.plot(x, y, 'k-', linewidth=0.5)
    for group, rows in df_group:
        x = rows['x'].values
        y = rows['y'].values
        plt.plot(x, y, 'k-', linewidth=2)
    plt.show()